/*
 * gshot for GP2X, r1
 * written by notaz, 2008
 *
 * This source code is dedicated to the public domain.
 * However, if you use this code and you are a nice person,
 * you may want to indicate that you used this code
 * in your program's documentation.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <errno.h>
#include <png.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <setjmp.h>
#include <unistd.h>
#include <signal.h>
#include "../common/gp2x.h"

#define RELEASE "1"

volatile unsigned short *gp2x_memregs = NULL;
static volatile unsigned int   *gp2x_memregl = NULL;
static int memdev = -1;

static unsigned char *screen = NULL;
static int pal[256];
static const int bpps[4] = { 4, 8, 16, 24 };
static png_bytepp row_pointers = NULL;

typedef struct { int ints[4]; } intstruct;

#define DEFAULT_PRIO -1


static int realloc_stuff(int w, int h, int stride, int bpp)
{
	static int screen_size = 0, rows = 0, cols = 0;
	int i, new_size, need_tbl_update = 0;

	new_size = stride*h*bpp/8;
	if (screen_size < new_size) {
		screen = realloc(screen, new_size);
		if (screen == NULL) goto oom;
		screen_size = new_size;
	}

	if (rows < h) {
		row_pointers = realloc(row_pointers, h * sizeof(row_pointers[0]));
		if (row_pointers == NULL) goto oom;
		memset(&row_pointers[rows], 0, (h - rows) * sizeof(row_pointers[0]));
		rows = h;
		need_tbl_update = 1;
	}

	if (cols < w || need_tbl_update) {
		for (i = 0; i < rows; i++) {
			row_pointers[i] = realloc(row_pointers[i], w*3);
			if (row_pointers[i] == NULL) goto oom;
		}
		cols = w;
	}

	return new_size;

oom:
	printf("gshot: OOM!\n");
	exit(1);
	return 0;
}

static void quickcpy(void *dst, void *src, int len)
{
	intstruct *d = dst;
	intstruct *s = src;

	for (len /= 16; len > 0; len--)
		*d++ = *s++;
}


static int dump_screen(int *w, int *h, int *stride)
{
	unsigned short msw1, msw2, lsw1, lsw2, bpp1, bpp2, scalew, scaleh;
	int addr, size, len, i, adjust;
	unsigned char *mem;

	i = setpriority(PRIO_PROCESS, 0, -10);
	if (i != 0) printf("setpriority failed: %i\n", i);

	msw1 = gp2x_memregs[0x2910>>1];
	msw2 = gp2x_memregs[0x2914>>1];
	lsw1 = gp2x_memregs[0x290E>>1];
	lsw2 = gp2x_memregs[0x2912>>1];

	// these should most often match
	if (msw1 != msw2)
		msw1 = gp2x_memregs[0x2914>>1];
	if (lsw1 != lsw2)
		lsw1 = gp2x_memregs[0x2912>>1];

	addr = ((int)msw1<<16) | lsw1;

	// bpp
	bpp1 = (gp2x_memregs[0x28DA>>1] >> 9) & 3; // 4, 8, 16, 24
	bpp2 = (gp2x_memregs[0x28DA>>1] >> 9) & 3;
	if (bpp1 != bpp2)
		bpp1 = (gp2x_memregs[0x28DA>>1] >> 9) & 3;

	// stride, scale
	*stride = gp2x_memregs[0x290c>>1]; // bytes
	scalew = gp2x_memregs[0x2906>>1];
	scaleh = gp2x_memregl[0x2908>>2];
	//printf("sw: %i, sh: %i\n", scalew, scaleh);

	if (scalew != 1024) { // scalew 1024 disables scaling completely
		printf("gshot: scaling detected\n");
		*h = (int)(240.0 * (double)scaleh / *stride + 0.5);
	}
	else *h = 240;

	*stride = bpp1 ? *stride/bpp1 : *stride*2; // make pixels

	// *w = (int)((double)*stride * scalew / 1024.0 + 0.5);
	*w = (int)((double)320.0 * scalew / 1024.0 + 0.5);

	size = realloc_stuff(*w, *h, *stride, bpps[bpp1]);

	adjust = addr & 0xfff;

	printf("gshot: about to dump %08x, %ix%i@%i, stride=%i, size=%i\n", addr, *w, *h, bpps[bpp1], *stride, size);

	mem = mmap(NULL, size + adjust, PROT_READ, MAP_SHARED, memdev, addr - adjust);
	if (mem == MAP_FAILED)
	{
		perror("gshot: mmap failed");
		setpriority(PRIO_PROCESS, 0, 0);
		return -1;
	}

	if (msw1 != gp2x_memregs[0x2910>>1] || msw2 != gp2x_memregs[0x2914>>1])
		printf("gshot: addr changed!\n");

	quickcpy(screen, mem + adjust, size);

	if (bpps[bpp1] < 16)
	{
		// must dump the palette too
		// let's hope the other program isn't writing palette now..
	        gp2x_memregs[0x2958>>1] = 0;

		for (len = 256/2, i = 0; len > 0; len--, i+=2)
		{
			pal[i+0]  = (int)gp2x_memregs[0x295a>>1];
			pal[i+0] |= (int)gp2x_memregs[0x295a>>1] << 16;
			pal[i+1]  = (int)gp2x_memregs[0x295a>>1];
			pal[i+1] |= (int)gp2x_memregs[0x295a>>1] << 16;
		}

	}

	munmap(mem, size + adjust);
	setpriority(PRIO_PROCESS, 0, DEFAULT_PRIO);

	return bpps[bpp1];
}


static void convert(int bpp, int w, int h, int stride)
{
	unsigned char *d;
	int x, y;

	switch (bpp)
	{
		case 4: {
			for (y = 0; y < h; y++)
			{
				unsigned char *s = (unsigned char *)screen + y*stride/2;
				d = row_pointers[y];
				for (x = 0; x < w*3; s++, x+=6)
				{
					int p = pal[*s & 0xf];
					d[x+0] = p >> 16;
					d[x+1] = p >>  8;
					d[x+2] = p;
					p = pal[*s >> 4];
					d[x+3] = p >> 16;
					d[x+4] = p >>  8;
					d[x+5] = p;
				}
			}
			break;
		}

		case 8: {
			for (y = 0; y < h; y++)
			{
				unsigned char *s = (unsigned char *)screen + y*stride;
				d = row_pointers[y];
				for (x = 0; x < w*3; s++, x+=3)
				{
					int p = pal[*s];
					d[x+0] = p >> 16;
					d[x+1] = p >>  8;
					d[x+2] = p;
				}
			}
			break;
		}

		case 16: {
			for (y = 0; y < h; y++)
			{
				unsigned short *s = (unsigned short *)screen + y*stride;
				d = row_pointers[y];
				for (x = 0; x < w*3; s++, x+=3)
				{
					d[x+0] = ((*s >> 11) * 264) >> 5; // *= 8.25
					d[x+1] = (((*s >> 5) & 0x3f) * 260) >> 6; // *= 4.0625
					d[x+2] = ((*s & 0x1f) * 264) >> 5;
				}
			}
			break;
		}

		case 24: {
			for (y = 0; y < h; y++)
			{
				unsigned char *s = (unsigned char *)screen + y*stride*3;
				d = row_pointers[y];
				for (x = 0; x < w*3; s+=3, x+=3)
				{
					d[x+0] = s[2];
					d[x+1] = s[1];
					d[x+2] = s[0];
				}
			}
			break;
		}

		default:
			printf("bad bpp: %i\n", bpp);
	}
}


static int write_png(const char *fname, int w, int h)
{
	png_structp png_ptr = NULL;
	png_infop info_ptr = NULL;
	FILE *f;

	f = fopen(fname, "wb");
	if (f == NULL) {
		printf("gshot: failed to open \"%s\"\n", fname);
		return -1;
	}

	/* initialize stuff */
	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

	if (png_ptr == NULL) {
		printf("png_create_write_struct() failed");
		goto fail1;
	}

	info_ptr = png_create_info_struct(png_ptr);
	if (info_ptr == NULL) {
		printf("png_create_info_struct() failed");
		goto fail2;
	}

	if (setjmp(png_jmpbuf(png_ptr)) != 0) {
		printf("error in png code\n");
		goto fail2;
	}

	png_init_io(png_ptr, f);

	png_set_IHDR(png_ptr, info_ptr, w, h,
		8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

	png_write_info(png_ptr, info_ptr);
	png_write_image(png_ptr, row_pointers);
	png_write_end(png_ptr, NULL);

        fclose(f);
	sync();
	return 0;

fail2:
	png_destroy_write_struct(&png_ptr, &info_ptr);
fail1:
	fclose(f);
	return -1;
}


/* interactive mode stuff */
static unsigned int keys[3] = { GP2X_START, GP2X_R, 0, };
static int signal_received = 0;

const char * const keyNames[] = {
	"UP",    "???",    "LEFT", "???",  "DOWN", "???", "RIGHT",    "???",
	"START", "SELECT", "L",    "R",    "A",    "B",   "X",        "Y",
	"???",   "???",    "???",  "???",  "???",  "???", "VOL DOWN", "VOL UP",
	"???",   "???",    "???",  "PUSH", "???",  "???", "???",      "???"
};


static const char *key_name(unsigned int mask)
{
	int i, u;

	if (!mask) return "<none>";
	for (u = 0, i = mask; 1; i >>= 1, u++)
		if (i & 1) break;

	if (keyNames[u][0] == '?') return NULL;
	return keyNames[u];
}


static char *get_next_fname(int new_group)
{
	static int fnum = 0;
	static char buff[32];
	FILE *f;

	while (1)
	{
		snprintf(buff, sizeof(buff), "shot%i_%04i.png", fnum>>16, fnum & 0xffff);
		f = fopen(buff, "r");
		if (f == NULL) break;
		fclose(f);
		fnum += new_group ? 0x10000 : 1;
	}

	return buff;
}

static void sigusr_handler(int signum)
{
	signal_received = 1;
}

static void install_signal_handler(void)
{
	struct sigaction sa;

	memset(&sa, 0, sizeof(sa));

	sa.sa_handler = sigusr_handler;
	sigaction(SIGUSR1, &sa, NULL);
}

int main(int argc, char *argv[])
{
	int bpp, ret = 0, w = 0, h = 0, stride = 0, nomenu = 0;
	char *fname;

  	memdev = open("/dev/mem", O_RDWR);
	if (memdev == -1)
	{
		printf("open(\"/dev/mem\") failed with %i\n", errno);
		ret = 1;
		goto end;
	}

	gp2x_memregs = mmap(0, 0x10000, PROT_READ|PROT_WRITE, MAP_SHARED, memdev, 0xc0000000);
	if (gp2x_memregs == MAP_FAILED)
	{
		printf("mmap(memregs) failed with %i\n", errno);
		ret = 2;
		goto end;
	}
	gp2x_memregl = (volatile unsigned int *)gp2x_memregs;

	if (argc > 1)
	{
		fname = argv[1];
		if (argc == 4 && strcmp(argv[1], "-s") == 0) {
			ret = atoi(argv[2]);
			printf("gshot: sleeping for %i secs\n", ret);
			sleep(ret);
			fname = argv[3];
		}
		else if (argc != 2) {
			printf("usage: %s [-s <seconds>] <filename>\n", argv[0]);
			printf("       or run without parameters for interactive mode\n");
			goto end;
		}
		bpp = dump_screen(&w, &h, &stride);
		convert(bpp, w, h, stride);
		ret = write_png(fname, w, h);
		if (ret != 0)
			printf("gshot: write_png() failed\n");
		else
			printf("gshot: done.\n");
	}
	else
	{
		// interactive mode
		unsigned char *scr;
		int sel = 0, pad, oldpad = 0, fbh_old, fbl_old;
		pid_t pid = 0;
		FILE *f;

		f = fopen("/tmp/gshot.pid", "r");
		if (f != NULL)
		{
			// maybe already running?
			fscanf(f, "%i", &pid);
			if (pid > 0 && kill(pid, SIGUSR1) == 0) {
				printf("other instance detected (pid %i), signaled and exiting\n", pid);
				nomenu = 1;
				goto end;
			}
			fclose(f);
		}

		f = fopen("gshot.bcfg", "rb");
		if (f != NULL) {
			fread(keys, 1, sizeof(keys), f);
			fclose(f);
		}

		fbh_old = gp2x_memregs[0x2914>>1];
		fbl_old = gp2x_memregs[0x2912>>1];

again:
		scr = mmap(NULL, 320*240, PROT_WRITE, MAP_SHARED, memdev, 0x02000000);
		if (scr == MAP_FAILED)
		{
			perror("mmap failed");
			goto end;
		}

		memset(scr, 0, 320*240);

	        gp2x_memregs[0x2958>>1] = 0; // pal
		gp2x_memregs[0x295a>>1] = 0; // b and w
		gp2x_memregs[0x295a>>1] = 0;
		gp2x_memregs[0x295a>>1] = 0xffff;
		gp2x_memregs[0x295a>>1] = 0x00ff;
		gp2x_memregs[0x28DA>>1] = (gp2x_memregs[0x28DA>>1] & ~(3<<9)) | (1<<9); // 8bpp
		gp2x_memregs[0x2906>>1] = 1024; // no scaling
		gp2x_memregs[0x290c>>1] = 320; // stride
		gp2x_memregs[0x2914>>1] = 0x0200; // dst
		gp2x_memregs[0x2912>>1] = 0x0000;

		text_out8(scr, 60,  35, "gshot r" RELEASE);
		text_out8(scr, 60,  45, "(c) notaz, 2008");
		text_out8(scr, 60,  68, "button combo:");
		text_out8(scr, 60, 150, "next file:");
		text_out8(scr, 60, 158, get_next_fname(1));

		while (1)
		{
			gp2x_video_wait_vsync();
			memset(scr + 320*76, 0, 320*6*8);
			text_out8(scr, 60,  76, key_name(keys[0]));
			text_out8(scr, 60,  84, key_name(keys[1]));
			text_out8(scr, 60,  92, key_name(keys[2]));
			text_out8(scr, 60, 108, "GO to background");
			text_out8(scr, 60, 116, "exit");
			text_out8(scr, 50, 76+sel*8, ">");

			while (1)
			{
				pad = gp2x_joystick_read();
				if        (pad & GP2X_UP) {
					sel--;
					if (sel <  0) sel = 5;
					if (sel == 3) sel = 2;
				} else if (pad & GP2X_DOWN) {
					sel++;
					if (sel >  5) sel = 0;
					if (sel == 3) sel = 4;
				} else if (pad & GP2X_LEFT) {
					if (sel < 3) {
						if (keys[sel] == 0) keys[sel] = 0x80000000;
						do {
							keys[sel] >>= 1;
						} while (key_name(keys[sel]) == NULL);
					}
				} else if (pad & GP2X_RIGHT) {
					if (sel < 3) {
						if (keys[sel] == 0) keys[sel] = 1;
						else do {
							keys[sel] <<= 1;
						} while (key_name(keys[sel]) == NULL);
					}
				} else if (pad & (GP2X_B|GP2X_START)) {
					if (sel == 4) goto endfork;
					if (sel == 5) goto endint;
					continue;
				} else
					continue;

				break;
			}
		}
endfork:
		munmap(scr, 320*240);
		gp2x_memregs[0x2914>>1] = fbh_old;
		gp2x_memregs[0x2912>>1] = fbl_old;

		f = fopen("gshot.bcfg", "wb");
		if (f != NULL) {
			fwrite(keys, 1, sizeof(keys), f);
			fclose(f);
		}

		pid = fork();
		if (pid == 0) // child
		{
			unsigned int mask;

			printf("gshot: child starting with pid %i\n", getpid());
			f = fopen("/tmp/gshot.pid", "w");
			if (f != NULL) {
				fprintf(f, "%i", getpid());
				fclose(f);
			}
			signal_received = 0;
			install_signal_handler();

			mask = keys[0]|keys[1]|keys[2];
			if (mask == 0) goto end;
			while (!signal_received)
			{
				pad = gp2x_joystick_read();
				if ((pad & mask) == mask && pad != oldpad)
				{
					fname = get_next_fname(0);
					bpp = dump_screen(&w, &h, &stride);
					convert(bpp, w, h, stride);
					ret = write_png(fname, w, h);
					if (ret != 0)
						printf("gshot: write_png() failed\n");
					else
						printf("gshot: wrote \"%s\".\n", fname);
				}
				oldpad = pad;
			}

			printf("gshot: signal received\n");
			remove("/tmp/gshot.pid");
			goto again;
		}
		goto end;

endint:
		munmap(scr, 320*240);
		gp2x_memregs[0x2914>>1] = fbh_old;
		gp2x_memregs[0x2912>>1] = fbl_old;
	}

end:
	if (gp2x_memregs != NULL)
		munmap((void *)gp2x_memregs, 0x10000);
	if (memdev != -1)
		close(memdev);
	if (argc <= 1 && !nomenu) {
		chdir("/usr/gp2x");
		execl("gp2xmenu", "gp2xmenu", NULL);
	}

	return ret;
}

